/****************************************************************************
 * Functions to handle loading colors.
 * Copyright (C) 2004 Joe Wingbermuehle
 ****************************************************************************/

#include "jwm.h"

/* Note: COLOR_HASH_SIZE = 2 ** (3 * COLOR_HASH_BITS) */
static const int COLOR_HASH_BITS = 3;
static const int COLOR_HASH_SIZE = 512;

typedef struct ColorNode {
	unsigned short red, green, blue;
	unsigned long pixel;
	struct ColorNode *prev;
	struct ColorNode *next;
} ColorNode;

static ColorNode **colorHash = NULL;

static const float COLOR_DELTA = 0.45;

unsigned long colors[COLOR_COUNT];
unsigned int rgbColors[COLOR_COUNT];

static int ParseColor(int index);
static unsigned int ReadHex(const char *hex);
static unsigned int GetRGBFromXColor(const XColor *c);
static void GetColor(XColor *c);
static int GetColorByName(const char *str, XColor *c);
static void InitializeNames();
static XColor LightenColor(XColor *old);
static XColor DarkenColor(XColor *old);
static void CreateColorRamp(unsigned int a, unsigned int b,
	unsigned long *ramp);

static Colormap cm;
static char **names = NULL;
unsigned long *ramps[RAMP_COUNT];

/****************************************************************************
 ****************************************************************************/
unsigned int GetRGBFromXColor(const XColor *c) {
	float red, green, blue;
	unsigned int rgb;

	red = (float)c->red / 65535.0;
	green = (float)c->green / 65535.0;
	blue = (float)c->blue / 65535.0;

	rgb = (int)(red * 255.0) << 16;
	rgb |= (int)(green * 255.0) << 8;
	rgb |= (int)(blue * 255.0);

	return rgb;
}

/****************************************************************************
 ****************************************************************************/
void InitializeColors() {
	XColor temp, temp2;
	int x;

	cm = DefaultColormap(display, DefaultScreen(display));

	colorHash = Allocate(sizeof(ColorNode*) * COLOR_HASH_SIZE);
	for(x = 0; x < COLOR_HASH_SIZE; x++) {
		colorHash[x] = NULL;
	}

	if(!ParseColor(COLOR_BORDER)) {

		GetColorByName("gray", &temp);
		colors[COLOR_BORDER] = temp.pixel;
		rgbColors[COLOR_BORDER] = GetRGBFromXColor(&temp);

		temp2 = LightenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_BORDER_UP] = temp2.pixel;
		rgbColors[COLOR_BORDER_UP] = GetRGBFromXColor(&temp2);

		temp2 = DarkenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_BORDER_DOWN] = temp2.pixel;
		rgbColors[COLOR_BORDER_DOWN] = GetRGBFromXColor(&temp2);

	}

	if(!ParseColor(COLOR_BORDER_OUTLINE)) {
		GetColorByName("black", &temp);
		colors[COLOR_BORDER_OUTLINE] = temp.pixel;
		rgbColors[COLOR_BORDER_OUTLINE] = GetRGBFromXColor(&temp);
	}

	if(!ParseColor(COLOR_BORDER_TEXT)) {
		GetColorByName("black", &temp);
		colors[COLOR_BORDER_TEXT] = temp.pixel;
		rgbColors[COLOR_BORDER_TEXT] = GetRGBFromXColor(&temp);
	}

	if(!ParseColor(COLOR_ACTIVE_BORDER)) {

		GetColorByName("blue", &temp);
		colors[COLOR_ACTIVE_BORDER] = temp.pixel;
		rgbColors[COLOR_ACTIVE_BORDER] = GetRGBFromXColor(&temp);

		temp2 = LightenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_ACTIVE_BORDER_UP] = temp2.pixel;
		rgbColors[COLOR_ACTIVE_BORDER_UP] = GetRGBFromXColor(&temp2);

		temp2 = DarkenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_ACTIVE_BORDER_DOWN] = temp2.pixel;
		rgbColors[COLOR_ACTIVE_BORDER_DOWN] = GetRGBFromXColor(&temp2);

	}

	if(!ParseColor(COLOR_ACTIVE_BORDER_TEXT)) {
		GetColorByName("white", &temp);
		colors[COLOR_ACTIVE_BORDER_TEXT] = temp.pixel;
		rgbColors[COLOR_ACTIVE_BORDER_TEXT] = GetRGBFromXColor(&temp);
	}

	if(!ParseColor(COLOR_BUTTON)) {

		GetColorByName("gray", &temp);
		colors[COLOR_BUTTON] = temp.pixel;
		rgbColors[COLOR_BUTTON] = GetRGBFromXColor(&temp);

		temp2 = LightenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_BUTTON_UP] = temp2.pixel;
		rgbColors[COLOR_BUTTON_UP] = GetRGBFromXColor(&temp2);

		temp2 = DarkenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_BUTTON_DOWN] = temp2.pixel;
		rgbColors[COLOR_BUTTON_DOWN] = GetRGBFromXColor(&temp2);

	}

	if(!ParseColor(COLOR_BUTTON_TEXT)) {
		GetColorByName("black", &temp);
		colors[COLOR_BUTTON_TEXT] = temp.pixel;
		rgbColors[COLOR_BUTTON_TEXT] = GetRGBFromXColor(&temp);
	}

	if(names) {
		for(x = 0; x < COLOR_COUNT; x++) {
			if(names[x]) {
				Release(names[x]);
			}
		}
		Release(names);
		names = NULL;
	}

	for(x = 0; x < RAMP_COUNT; x++) {
		ramps[x] = Allocate(256 * sizeof(unsigned long));
	}

	CreateColorRamp(rgbColors[COLOR_BORDER], rgbColors[COLOR_BORDER_TEXT],
		ramps[RAMP_INACTIVE]);

	CreateColorRamp(rgbColors[COLOR_ACTIVE_BORDER],
		rgbColors[COLOR_ACTIVE_BORDER_TEXT],
		ramps[RAMP_ACTIVE]);

	CreateColorRamp(rgbColors[COLOR_BUTTON], rgbColors[COLOR_BUTTON_TEXT],
		ramps[RAMP_BUTTON]);

}

/****************************************************************************
 ****************************************************************************/
void ReleaseColors() {
	ColorNode *np;
	int x;

	if(names) {
		for(x = 0; x < COLOR_COUNT; x++) {
			if(names[x]) {
				Release(names[x]);
			}
		}
		Release(names);
		names = NULL;
	}

	for(x = 0; x < RAMP_COUNT; x++) {
		Release(ramps[x]);
	}

	if(colorHash) {

		for(x = 0; x < COLOR_HASH_SIZE; x++) {
			while(colorHash[x]) {
				np = colorHash[x]->next;
				XFreeColors(display, cm, &colorHash[x]->pixel, 1, 0);
				Release(colorHash[x]);
				colorHash[x] = np;
			}
		}

		Release(colorHash);
	}
}

/****************************************************************************
 ****************************************************************************/
void SetTextColor(const char *value) {

	Assert(value);

	InitializeNames();
	names[COLOR_BORDER_TEXT] = Allocate(strlen(value) + 1);
	strcpy(names[COLOR_BORDER_TEXT], value);
}

/****************************************************************************
 ****************************************************************************/
void SetActiveTextColor(const char *value) {

	Assert(value);

	InitializeNames();
	names[COLOR_ACTIVE_BORDER_TEXT] = Allocate(strlen(value) + 1);
	strcpy(names[COLOR_ACTIVE_BORDER_TEXT], value);
}

/****************************************************************************
 ****************************************************************************/
void SetBorderColor(const char *value) {

	Assert(value);

	InitializeNames();
	names[COLOR_BORDER] = Allocate(strlen(value) + 1);
	strcpy(names[COLOR_BORDER], value);
}

/****************************************************************************
 ****************************************************************************/
void SetActiveBorderColor(const char *value) {

	Assert(value);

	InitializeNames();
	names[COLOR_ACTIVE_BORDER] = Allocate(strlen(value) + 1);
	strcpy(names[COLOR_ACTIVE_BORDER], value);
}

/****************************************************************************
 ****************************************************************************/
void SetButtonColor(const char *value) {

	Assert(value);

	InitializeNames();
	names[COLOR_BUTTON] = Allocate(strlen(value) + 1);
	strcpy(names[COLOR_BUTTON], value);
}

/****************************************************************************
 ****************************************************************************/
void SetButtonTextColor(const char *value) {

	Assert(value);

	InitializeNames();
	names[COLOR_BUTTON_TEXT] = Allocate(strlen(value) + 1);
	strcpy(names[COLOR_BUTTON_TEXT], value);
}

/****************************************************************************
 ****************************************************************************/
int ParseColor(int index) {
	XColor temp, temp2;
	unsigned int rgb;

	if(!names || !names[index]) {
		return 0;
	}

	if(names[index][0] == '#' && strlen(names[index]) == 7) {
		rgb = ReadHex(names[index] + 1);
		temp.red = ((rgb >> 16) & 0xFF) * 257;
		temp.green = ((rgb >> 8) & 0xFF) * 257;
		temp.blue = (rgb & 0xFF) * 257;
		temp.flags = DoRed | DoGreen | DoBlue;
		GetColor(&temp);
	} else {
		if(!GetColorByName(names[index], &temp)) {
			Warning("bad color: \"%s\"", names[index]);
			return 0;
		}
	}
	colors[index] = temp.pixel;
	rgbColors[index] = GetRGBFromXColor(&temp);

	if(index == COLOR_BORDER) {

		temp2 = LightenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_BORDER_UP] = temp2.pixel;
		rgbColors[COLOR_BORDER_UP] = GetRGBFromXColor(&temp2);

		temp2 = DarkenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_BORDER_DOWN] = temp2.pixel;
		rgbColors[COLOR_BORDER_DOWN] = GetRGBFromXColor(&temp2);

	} else if(index == COLOR_ACTIVE_BORDER) {

		temp2 = LightenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_ACTIVE_BORDER_UP] = temp2.pixel;
		rgbColors[COLOR_ACTIVE_BORDER_UP] = GetRGBFromXColor(&temp2);

		temp2 = DarkenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_ACTIVE_BORDER_DOWN] = temp2.pixel;
		rgbColors[COLOR_ACTIVE_BORDER_DOWN] = GetRGBFromXColor(&temp2);

	} else if(index == COLOR_BUTTON) {

		temp2 = LightenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_BUTTON_UP] = temp2.pixel;
		rgbColors[COLOR_BUTTON_UP] = GetRGBFromXColor(&temp2);

		temp2 = DarkenColor(&temp);
		GetColor(&temp2);
		colors[COLOR_BUTTON_DOWN] = temp2.pixel;
		rgbColors[COLOR_BUTTON_DOWN] = GetRGBFromXColor(&temp2);

	}

	return 1;
}

/****************************************************************************
 ****************************************************************************/
void InitializeNames() {
	int x;

	if(names == NULL) {
		names = Allocate(sizeof(char*) * COLOR_COUNT);
		for(x = 0; x < COLOR_COUNT; x++) {
			names[x] = NULL;
		}
	}
}

/****************************************************************************
 ****************************************************************************/
unsigned int ReadHex(const char *hex) {
	unsigned int value = 0;
	unsigned int x;

	for(x = 0; hex[x]; x++) {
		value *= 16;
		if(hex[x] >= '0' && hex[x] <= '9') {
			value += hex[x] - '0';
		} else if(hex[x] >= 'A' && hex[x] <= 'F') {
			value += hex[x] - 'A' + 10;
		} else if(hex[x] >= 'a' && hex[x] <= 'f') {
			value += hex[x] - 'a' + 10;
		}
	}

	return value;
}

/****************************************************************************
 ****************************************************************************/
XColor LightenColor(XColor *old) {
	XColor ret = *old;

	float red, green, blue;
	float delta = 1.0 + COLOR_DELTA;

	red = (float)ret.red / 65535.0;
	green = (float)ret.green / 65535.0;
	blue = (float)ret.blue / 65535.0;

	red = Min(delta * red, 1.0);
	green = Min(delta * green, 1.0);
	blue = Min(delta * blue, 1.0);

	ret.red = red * 65535.0;
	ret.green = green * 65535.0;
	ret.blue = blue * 65535.0;

	return ret;
}

/****************************************************************************
 ****************************************************************************/
XColor DarkenColor(XColor *old) {
	XColor ret = *old;

	float red, green, blue;
	float delta = 1.0 - COLOR_DELTA;

	red = (float)ret.red / 65535.0;
	green = (float)ret.green / 65535.0;
	blue = (float)ret.blue / 65535.0;

	red = delta * red;
	green = delta * green;
	blue = delta * blue;

	ret.red = red * 65535.0;
	ret.green = green * 65535.0;
	ret.blue = blue * 65535.0;

	return ret;
}

/***************************************************************************
 ***************************************************************************/
void CreateColorRamp(unsigned int a, unsigned int b, unsigned long *ramp) {

	XColor color;
	unsigned int x;
	float value, inv;

	float reda = (float)((a >> 16) & 0xFF) / 255.0;
	float greena = (float)((a >> 8) & 0xFF) / 255.0;
	float bluea = (float)(a & 0xFF) / 255.0;

	float redb = (float)((b >> 16) & 0xFF) / 255.0;
	float greenb = (float)((b >> 8) & 0xFF) / 255.0;
	float blueb = (float)(b & 0xFF) / 255.0;

	Assert(ramp);

	for(x = 0; x < 256; x++) {

		value = (float)x / 255.0;
		inv = 1.0 - value;

		color.red = (int)((reda * inv + redb * value) * 65535.0 + 0.5);
		color.green = (int)((greena * inv + greenb * value) * 65535.0 + 0.5);
		color.blue = (int)((bluea * inv + blueb * value) * 65535.0 + 0.5);

		GetColor(&color);
		ramp[x] = color.pixel;

	}

}

/***************************************************************************
 ***************************************************************************/
int GetColorByName(const char *str, XColor *c) {

	Assert(str);
	Assert(c);

	if(!XParseColor(display, cm, str, c)) {
		return 0;
	}

	GetColor(c);

	return 1;

}

/***************************************************************************
 ***************************************************************************/
void GetColor(XColor *c) {
	ColorNode *np;
	int hash;

	Assert(c);

	hash = c->red >> (8 + 5 - COLOR_HASH_BITS * 2);
	hash |= c->green >> (8 + 5 - COLOR_HASH_BITS);
	hash |= c->blue >> (8 + 5);

	Assert(hash >= 0);
	Assert(hash < COLOR_HASH_SIZE);

	np = colorHash[hash];
	while(np) {
		if(c->red == np->red && c->green == np->green && c->blue == np->blue) {
			if(np->prev) {
				np->prev->next = np->next;
				if(np->next) {
					np->next->prev = np->prev;
				}
				np->prev = NULL;
				colorHash[hash]->prev = np;
				np->next = colorHash[hash];
				colorHash[hash] = np;
			}
			c->pixel = np->pixel;
			return;
		}
		np = np->next;
	}

	np = Allocate(sizeof(ColorNode));

	np->red = c->red;
	np->green = c->green;
	np->blue = c->blue;

	if(!XAllocColor(display, cm, c)) {
		Warning("could not allocate color: %4X %4X %4X",
			c->red, c->green, c->blue);
	}
	np->pixel = c->pixel;

	if(colorHash[hash]) {
		colorHash[hash]->prev = np;
	}
	np->prev = NULL;
	np->next = colorHash[hash];
	colorHash[hash] = np;

}

